.xed-window .primary-toolbar {
    border: 0px solid @borders;
    border-bottom-width: 1px;
}

.xed-statusbar {
    border: 0px solid @borders;
    border-top-width: 1px;
}

.xed-searchbar {
    box-shadow: inset 0 1px @borders;
}

.xed-statusbar button {
    padding: 0 4px;
}
