.xed-map-frame:dir(ltr) border {
    border-width: 0 0 0 1px;
}

.xed-map-frame:dir(rtl) border {
    border-width: 0 1px 0 0;
}

.xed-searchbar .close-button {
    padding: 0;
    min-height: 0;
    min-width: 0;
}

/*remove a keybinding so that we can use it for toggling comments*/
@binding-set xed-view-bindings
{
    unbind "<Control>slash";
}

textview
{
    -gtk-key-bindings: xed-view-bindings;
}
